/*-
 * Copyright (c) 1988, 1989 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * From: @(#)rshd.c	5.38 (Berkeley) 3/2/91
 */

/* rrshd.c, derived from netkit-0.17 rshd (v1.25) by Paul A. Whittaker
 * <Paul.Whittaker@iname.com>, June 2002.
 *
 * rrshd (Restricted Remote Shell Daemon) is a cut-down, standalone version of
 * rshd.  It allows only specific commands to be run, and only as a set user
 * (nobody).  It pays no attention to local or remote usernames, /etc/nologin,
 * .rhosts or hosts.equiv files.
 *
 * Version: 1.2
 *
 * 25 Aug 2012
 * - add <linux/limits.h> for ARG_MAX.
 * 17 Jul 2003
 * - read RRSH_ALLOW file *before* dropping privilege, not after, so that its
 * permissions can be limited to prevent the user from discovering which
 * commands are valid.
 *
 * Version: 1.1
 *
 * 12 Jun 2003
 * - added support for SHELL, PATH, and HOME overrides, and optional setting
 * of a fixed DISPLAY value.
 *
 * Version: 1.0
 *
 * 23 Oct 2002
 * - modified LIBWRAP code to use refuse() instead of merely closing the
 * connection (adds a delay to prevent DoS attacks).
 * 16 Jun 2002
 * - initial code adaptation from netkit-0.17.
 */

/*
 * remote shell server:
 *	[port]\0
 *	remuser\0
 *	locuser\0
 *	command\0
 *	data
 */
#include <linux/limits.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <signal.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <pwd.h>
#include <grp.h>
#include <syslog.h>
#include <resolv.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>  /* for vsnprintf */
#include <stdlib.h>
#include <string.h>
#include <paths.h>
#include <stdarg.h>
#include <ctype.h>
#include <assert.h>
#ifdef LIBWRAP
#include <tcpd.h>
#endif

#define	OPTIONS	"Ln"

#ifndef RRSH_USER
#define RRSH_USER "nobody"
#endif
#ifndef RRSH_ALLOW
#define RRSH_ALLOW "/etc/rrsh.allow"
#endif
#ifndef RRSH_PATH
#define RRSH_PATH _PATH_DEFPATH
#endif
/* If you don't want to trust /etc/passwd, you may set these: */
/* #define RRSH_HOME "/tmp" */
/* #define RRSH_SHELL "/bin/sh" */
/* Optionally, you may wish to specify a default X11 display: */
/* #define RRSH_DISPLAY ":0" */

static int keepalive = 1;
static int beparanoid = 1;
static int sent_null;
#ifdef LIBWRAP
int allow_severity = LOG_INFO;
int deny_severity = LOG_WARNING;
#endif

char	username[20] = "USER=";
char	homedir[64] = "HOME=";
char	shell[64] = "SHELL=";
char	path[100] = "PATH=";
char	remoteuser[30] = "REMOTEUSER=";
char	remotehost[64] = "REMOTEHOST=";
#if defined(RRSH_DISPLAY)
char	display[64] = "DISPLAY=";
#endif
char	*envinit[] =
	    {homedir, shell, path, username, remoteuser, remotehost,
#if defined(RRSH_DISPLAY)
	    display,
#endif
	    /* general-purpose hook for shell environment */
	    "ENV=/etc/profile",
	    0};
extern	char	**environ;

static void error(const char *fmt, ...);
static void doit(struct sockaddr_in *fromp);
static void getstr(char *buf, int cnt, const char *err);

/*
 * Report error to client.
 * Note: can't be used until second socket has connected
 * to client, or older clients will hang waiting
 * for that connection first.
 */
static void
error(const char *fmt, ...) {
    va_list ap;
    char buf[BUFSIZ], *bp = buf;
    
    if (sent_null == 0)	*bp++ = 1;
    va_start(ap, fmt);
    vsnprintf(bp, sizeof(buf)-1, fmt, ap);
    va_end(ap);
    write(2, buf, strlen(buf));
}

static void fail(const char *errorstr, 
		 const char *remuser, const char *hostname, 
		 const char *cmdbuf) 
{
	/* log the (failed) rsh request */
	syslog(LOG_INFO|LOG_AUTH, "rsh denied to %s@%s: %s",
	       remuser, hostname, errorstr);
	if (beparanoid) {
	    syslog(LOG_INFO|LOG_AUTH, "rsh command was '%s'", cmdbuf);
	}
	error(errorstr, hostname);
}

static void getstr(char *buf, int cnt, const char *err) {
    char c;
    do {
	if (read(0, &c, 1) != 1) exit(1);
	*buf++ = c;
	if (--cnt == 0) {
	    error("%s too long\n", err);
	    exit(1);
	}
    } while (c != 0);
}

static int getint(void) {
    int port = 0;
    char c;
    do {
	if (read(0, &c, 1) != 1) exit(1);
	if (isascii(c) && isdigit(c)) port = port*10 + c-'0';
    } while (c != 0);
    return(port);
}

static void stderr_parent(int sock, int pype, int pid) {
    fd_set ready, readfrom;
    char buf[BUFSIZ], sig;
    int one = 1;
    int nfd, cc, guys=2;
    
    ioctl(pype, FIONBIO, (char *)&one);
    /* should set s nbio! */
    
    FD_ZERO(&readfrom);
    FD_SET(sock, &readfrom);
    FD_SET(pype, &readfrom);
    if (pype > sock) nfd = pype+1;
    else nfd = sock+1;
    
    while (guys > 0) {
	ready = readfrom;
	if (select(nfd, &ready, NULL, NULL, NULL) < 0) {
	   if (errno != EINTR) {
	      break;
	   }
	   continue;
	}
	if (FD_ISSET(sock, &ready)) {
	    cc = read(sock, &sig, 1);
	    if (cc <= 0) {
	       FD_CLR(sock, &readfrom);
	       guys--;
	    }
	    else killpg(pid, sig);
	}
	if (FD_ISSET(pype, &ready)) {
	    cc = read(pype, buf, sizeof(buf));
	    if (cc <= 0) {
		shutdown(sock, 2);
		FD_CLR(pype, &readfrom);
		guys--;
	    } 
	    else write(sock, buf, cc);
	}
    }
}


static const char *findhostname(struct sockaddr_in *fromp,
				const char *remuser, const char *cmdbuf) 
{
	struct hostent *hp;
	const char *hostname;

	hp = gethostbyaddr((char *)&fromp->sin_addr, sizeof (struct in_addr),
			   fromp->sin_family);

	errno = ENOMEM; /* malloc (thus strdup) may not set it */
	if (hp) hostname = strdup(hp->h_name);
	else hostname = strdup(inet_ntoa(fromp->sin_addr));

	if (hostname==NULL) {
	    /* out of memory? */
	    error("strdup: %s\n", strerror(errno));
	    exit(1);
	}

	/*
	 * Attempt to confirm the DNS. 
	 */
#ifdef	RES_DNSRCH
	_res.options &= ~RES_DNSRCH;
#endif
	hp = gethostbyname(hostname);
	if (hp == NULL) {
	    syslog(LOG_INFO, "Couldn't look up address for %s", hostname);
	    fail("Couldn't get address for your host (%s)\n", 
		 remuser, inet_ntoa(fromp->sin_addr), cmdbuf);
	    exit(1);
	} 
	while (hp->h_addr_list[0] != NULL) {
	    if (!memcmp(hp->h_addr_list[0], &fromp->sin_addr,
			sizeof(fromp->sin_addr))) {
		return(hostname);
	    }
	    hp->h_addr_list++;
	}
	syslog(LOG_NOTICE, "Host addr %s not listed for host %s",
	       inet_ntoa(fromp->sin_addr), hp->h_name);
	fail("Host address mismatch for %s\n", 
	     remuser, inet_ntoa(fromp->sin_addr), cmdbuf);
	exit(1);
}

static char **
checkcmd(const char *cmdbuf)

/* This function performs two tasks.  First, it constructs an argument vector
 * from a collapsed string, interpreting quotes (', ", \) in accordance with
 * shell grammar (except that variable substitution and backtick evaluation
 * are not performed).  It then verifies that the first argument (the command
 * name) literally matches a line in the allowed commands file, or the last
 * component of a line.  If any problem occurs, a null vector is returned to
 * indicate that the command is invalid or unauthorised.
 */

{
	static char buffer[(ARG_MAX+1)*2];
	static char *cmdvec[32];
	int vtop = 0, btop = 0, opensgl = 0, opendbl = 0, quote = 0, i;
	char *c = (char *) cmdbuf - 1, *token = buffer, linebuf[128];
	FILE *cf;

	bzero(buffer,sizeof(buffer));
	bzero(cmdvec,sizeof(cmdvec));

	while (vtop < 31 && btop < sizeof(buffer)-1 && *(++c) != 0) {
		switch (*c) {
		case '\'':
			if (opendbl || quote)
				break;
			opensgl = !opensgl;
			continue;
		case '"':
			if (opensgl || quote)
				break;
			opendbl = !opendbl;
			continue;
		case '\\':
			if (opensgl || quote)
				break;
			if (opendbl) {
				switch (*(c+1)) {
				case '$':
				case '`':
				case '\\':
				case '\n':
					quote = 1;
					continue;
				}
				break;
			}
			quote = 1;
			continue;
		case ' ':
		case '\t':
		case '\n':
			if (quote && *c == '\n') {
				quote = 0;
				continue;
			}
			if (opensgl || opendbl || quote)
		       		break;
			cmdvec[vtop++] = token;
			token = buffer + ++btop;
			continue;
		}
		quote = 0;
		buffer[btop++] = *c;
	}
	if (opensgl || opendbl || quote) {
		error("Syntax error.\n");
		exit(1);
	}
	if ((vtop == 31) || (btop == sizeof(buffer)-1)) {
		/*
		 * Might be short-changing the user a few bytes here; life is
		 * tough.
		 */
		syslog(LOG_ERR, "argument vector overflow");
		error("Command too long.\n");
		exit(1);
	}
	cmdvec[vtop] = token;

	cf = fopen(RRSH_ALLOW, "r");
	if (cf == NULL) {
		syslog(LOG_ERR, "commands file %s inaccessable", RRSH_ALLOW);
		error("Unauthorized command.\n");
		exit(1);
	}
	while (1) {
		if (fgets(linebuf, sizeof(linebuf), cf) == NULL)
			break;
		if (linebuf[0] == '#')
			continue;
		i = strlen(linebuf)-1;
		if (linebuf[i] == '\n')
			linebuf[i] = 0;
		if (!strcmp(cmdvec[0], linebuf)) {
			fclose(cf);
			return(cmdvec);
		}
		token = strrchr(linebuf, '/');
		if ((token != NULL) && !strcmp(cmdvec[0], ++token)) {
			i = sizeof(buffer) - ++btop;
			/*
			 * N.B. the above bounds checks guarantee that i>0
			 * here.
			 */
			strncpy(buffer + btop, linebuf, i);
			cmdvec[0] = buffer + btop;
			fclose(cf);
			return(cmdvec);
		}
	}
	fclose(cf);
	return(NULL);
}

static void
doit(struct sockaddr_in *fromp)
{
	char cmdbuf[ARG_MAX+1];
	char locuser[16], remuser[16];
	struct passwd *pwd;
	int sock = -1;
	const char *hostname;
	u_short port;
	int pv[2], pid, ifd;
	char **safecmd;

	signal(SIGINT, SIG_DFL);
	signal(SIGQUIT, SIG_DFL);
	signal(SIGTERM, SIG_DFL);
	signal(SIGCHLD, SIG_DFL);

	alarm(60);
	port = getint();
	alarm(0);

	if (port != 0) {
		int lport = IPPORT_RESERVED - 1;
		sock = rresvport(&lport);
		if (sock < 0) {
		    syslog(LOG_ERR, "can't get stderr port: %m");
		    exit(1);
		}
		if (port >= IPPORT_RESERVED) {
		    syslog(LOG_ERR, "2nd port not reserved\n");
		    exit(1);
		}
		fromp->sin_port = htons(port);
		if (connect(sock, (struct sockaddr *)fromp,
			    sizeof(*fromp)) < 0) {
		    syslog(LOG_INFO, "connect second port: %m");
		    exit(1);
		}
	}

	getstr(remuser, sizeof(remuser), "remuser");
	getstr(locuser, sizeof(locuser), "locuser");
	getstr(cmdbuf, sizeof(cmdbuf), "command");

	hostname = findhostname(fromp, remuser, cmdbuf);
	strncat(remotehost, hostname, sizeof(remotehost)-12);
	remotehost[sizeof(remotehost)-1] = 0;

	setpwent();
	pwd = getpwnam(RRSH_USER);
	if (pwd == NULL) {
		fail("Permission denied.\n", 
		     remuser, hostname, cmdbuf);
		syslog(LOG_ERR, "no such user (%s)", RRSH_USER);
		exit(1);
	}
	(void) write(2, "\0", 1);
	sent_null = 1;

	if (port) {
		if (pipe(pv) < 0) {
			error("Can't make pipe.\n");
			exit(1);
		}
		pid = fork();
		if (pid == -1)  {
			error("Can't fork; try again.\n");
			exit(1);
		}
		if (pid) {
			close(0); 
			close(1);
			close(2); 
			close(pv[1]);
			stderr_parent(sock, pv[0], pid);
			waitpid (pid, NULL, 0);
			exit(0);
		}
		setpgrp();
		close(sock); 
		close(pv[0]);
		dup2(pv[1], 2);
		close(pv[1]);
	}

	safecmd = checkcmd(cmdbuf);
	if (safecmd == NULL) {
		error("Unauthorized command.\n");
		exit(1);
	}

	if (setgid(pwd->pw_gid)) {
		syslog(LOG_ERR, "setgid: %m");
		exit(1);
	}
	if (initgroups(pwd->pw_name, pwd->pw_gid)) {
		syslog(LOG_ERR, "initgroups: %m");
		exit(1);
	}
	if (setuid(pwd->pw_uid)) {
		syslog(LOG_ERR, "setuid: %m");
		exit(1);
	}
	environ = envinit;

#if defined(RRSH_HOME)
	strncat(homedir, RRSH_HOME, sizeof(homedir)-6);
#else
	strncat(homedir, pwd->pw_dir, sizeof(homedir)-6);
#endif
	homedir[sizeof(homedir)-1] = 0;

	strncat(path, RRSH_PATH, sizeof(path)-6);
	path[sizeof(path)-1] = 0;

#if defined(RRSH_SHELL)
	strncat(shell, RRSH_SHELL, sizeof(shell)-7);
#else
	strncat(shell, pwd->pw_shell, sizeof(shell)-7);
#endif
	shell[sizeof(shell)-1] = 0;

	strncat(username, pwd->pw_name, sizeof(username)-6);
	username[sizeof(username)-1] = 0;

	strncat(remoteuser, remuser, sizeof(remoteuser)-12);
	remoteuser[sizeof(remoteuser)-1] = 0;

#if defined(RRSH_DISPLAY)
	strncat(display, RRSH_DISPLAY, sizeof(display)-9);
	display[sizeof(display)-1] = 0;
#endif

	endpwent();
	if (beparanoid) {
	    syslog(LOG_INFO|LOG_AUTH, "%s@%s: cmd='%s'",
		   remuser, hostname, cmdbuf);
	}

	/*
	 * Close all fds, in case libc has left fun stuff like 
	 * /etc/shadow open.
	 */
	for (ifd = getdtablesize()-1; ifd > 2; ifd--) close(ifd);

	if (chdir(homedir) < 0) {
		chdir("/");
		/*
		 * error("No remote directory.\n");
		 * exit(1);
		 */
	}

	execv(safecmd[0], safecmd);
	perror(safecmd[0]);
	exit(1);
}

static void network_init(int fd, struct sockaddr_in *fromp)
{
	struct linger linger;
	socklen_t fromlen;
	int on=1;
	int port;
#ifdef LIBWRAP
	struct request_info req;
#endif

	fromlen = sizeof(*fromp);
	if (getpeername(fd, (struct sockaddr *) fromp, &fromlen) < 0) {
		syslog(LOG_ERR, "getpeername: %m");
		_exit(1);
	}
	if (keepalive &&
	    setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, (char *)&on,
	    sizeof(on)) < 0)
		syslog(LOG_WARNING, "setsockopt (SO_KEEPALIVE): %m");
	linger.l_onoff = 1;
	linger.l_linger = 60;
	if (setsockopt(fd, SOL_SOCKET, SO_LINGER, (char *)&linger,
	    sizeof (linger)) < 0)
		syslog(LOG_WARNING, "setsockopt (SO_LINGER): %m");

	if (fromp->sin_family != AF_INET) {
	    syslog(LOG_ERR, "malformed \"from\" address (af %d)\n",
		   fromp->sin_family);
	    exit(1);
	}
#ifdef IP_OPTIONS
      {
	u_char optbuf[BUFSIZ/3], *cp;
	char lbuf[BUFSIZ+1], *lp;
	socklen_t optsize = sizeof(optbuf);
	int  ipproto;
	struct protoent *ip;

	if ((ip = getprotobyname("ip")) != NULL)
		ipproto = ip->p_proto;
	else
		ipproto = IPPROTO_IP;
	if (!getsockopt(0, ipproto, IP_OPTIONS, (char *)optbuf, &optsize) &&
	    optsize != 0) {
		lp = lbuf;

		/*
		 * If these are true, this will not run off the end of lbuf[].
		 */
		assert(optsize <= BUFSIZ/3);
		assert(3*optsize <= BUFSIZ);
		for (cp = optbuf; optsize > 0; cp++, optsize--, lp += 3)
			snprintf(lp, 4, " %2.2x", *cp);

		syslog(LOG_NOTICE,
		       "Connection received from %s using IP options"
		       " (ignored): %s",
		       inet_ntoa(fromp->sin_addr), lbuf);

		if (setsockopt(0, ipproto, IP_OPTIONS, NULL, optsize) != 0) {
			syslog(LOG_ERR, "setsockopt IP_OPTIONS NULL: %m");
			exit(1);
		}
	}
      }
#endif

	/*
	 * Check originating port for validity.
	 */
	port = ntohs(fromp->sin_port);
	if (port >= IPPORT_RESERVED || port < IPPORT_RESERVED/2) {
	    syslog(LOG_NOTICE|LOG_AUTH, "Connection from %s on illegal port",
		   inet_ntoa(fromp->sin_addr));
	    exit(1);
	}

#ifdef LIBWRAP
	/*
	 * Consult TCP wrappers
	 */
	request_init(&req, RQ_DAEMON, "rrshd", RQ_FILE, fd, NULL);
	fromhost(&req);
	if (!hosts_access(&req)) {
	    syslog(LOG_WARNING|LOG_AUTH,
			    "Connection from %s refused by tcp_wrappers",
			    eval_client(&req));
	    refuse(&req);
	    exit(1);
	}
	syslog(LOG_INFO|LOG_AUTH, "Connection from %s granted by tcp_wrappers",
		eval_client(&req));
#endif
}

void
reap()
{
	int status;

	wait(&status);
	signal(SIGCHLD, &reap);
}

int
main(int argc, char *argv[])
{
	int ch, sock, fd, yes=1, socksize;
	struct sockaddr_in from;
	struct protoent *prot;
	struct sockaddr_in addr;
	struct servent *serv;

	opterr = 0;
	while ((ch = getopt(argc, argv, OPTIONS)) != EOF) {
		switch (ch) {
		case 'n':
			keepalive = 0;
			break;

		case 'L':
			beparanoid = 0;
			break;

		case '?':
		default:
			syslog(LOG_ERR, "usage: rshd [-%s]", OPTIONS);
			exit(2);
		}
	}
	argc -= optind;
	argv += optind;

	freopen("/dev/null", "r", stdin);
	freopen("/dev/null", "w", stdout);
	freopen("/dev/null", "w", stderr);
	switch (fork()) {
	case -1:
		syslog(LOG_ERR, "fork: %m");
		exit(1);
	case 0:
		break;
	default:
		return(0);
	}
	setsid();
	openlog("rrshd", LOG_PID | LOG_ODELAY, LOG_DAEMON);
	signal(SIGCHLD, &reap);

	prot = getprotobyname("tcp");
	if (prot == NULL)
		sock = socket(PF_INET,SOCK_STREAM,IPPROTO_TCP);
	else
		sock = socket(PF_INET,SOCK_STREAM,prot -> p_proto);
	if (sock < 0) {
		syslog(LOG_ERR, "socket: %m");
		exit(1);
	}
	if (setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,&yes,sizeof(yes)) < 0)
		syslog(LOG_WARNING, "setsockopt (SO_REUSEADDR): %m");
	serv = getservbyname("shell", "tcp");
	if (serv == NULL)
		addr.sin_port = 514;
	else
		addr.sin_port = serv->s_port;
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
	if (bind(sock,(struct sockaddr *) &addr,sizeof(addr)) < 0) {
		syslog(LOG_ERR, "bind: %m");
		exit(1);
	}
	if (listen(sock,SOMAXCONN) < 0) {
		syslog(LOG_ERR, "listen: %m");
		exit(1);
	}

	socksize = sizeof(from);
 	while ((fd = accept(sock,(struct sockaddr *) &from,&socksize)) != -1) {
		switch (fork()) {
		case -1:
			syslog(LOG_ERR, "fork: %m");
			exit(1);
		case 0:
			network_init(fd, &from);
			dup2(fd, 0);
			dup2(fd, 1);
			dup2(fd, 2);
			close(fd);
			doit(&from);
		}
		close(fd);
	}
	syslog(LOG_ERR, "accept: %m");
	close(sock);
	exit(1);
}
